#donotrun
#info 3D-PK-1.0.5 by Patryk Kizny
#info Original shader by syntopia. Modifications by knight

/*
Simple 3D setup with anti-alias and DOF.

Implement this function in fragments that include this file:
vec3 color(vec3 cameraPos, vec3 direction);
vec3 backgroundMap (vec2 pos);
vec3 backgroundColor (vec3 pos);

Optional triggers
#define providesBackground
#define providesBackgroundMap
#define providesSRCurrentPt
#define providesInit

// -- Log
// Assumes square shaped pixel -> no over-blurring along x axis if the rendering window width is small wrt its height
// Added polygon shaped diaphragm.
// Added Control for the width of the in-focus region.

// -- Patryk Kizny Mods 
// Added More Intuitive FOV (horizontal, degrees)
// Added More Intuitive FOV (based on film gate height and focal length)
// Added Extra set of vectors Ey_ and Targe_ for animation of the camera
// 		these are stacked onto the base camera settings, but can be animated via scripting or tweening.
// Added camera readout from external data
// Added background 2D
// Changes along with 1.0.8 Raytracer - background removed from Post processing, Post group moved from BufferShader to 3D
*/

#buffer RGBA32F
#buffershader "BufferShader-1.0.2.frag"
#camera 3D
#vertex

#group Camera
// Field-of-view (perspective)
uniform float FOV; slider[0,1.35,2.0] NotLockable

// Field-of-view (horizontal in degrees)
uniform bool UseTrueFOV; checkbox[true]
uniform float TrueFOV; slider[0.1,115,179]

// Field-of-view (calculated based on film gate and focal length)
uniform bool UseFocalLength; checkbox[false]
uniform float FocalLength; slider[1,14,800]
uniform float FilmGate; slider[1,36,120]

varying float FOVP; // perspective FOV (0-2)
varying float FOVT; // Horizontal FOV in degrees
varying float FOVFL; // FOV using filmgate and focal length
uniform vec3 Eye; slider[(-50,-50,-50),(0,-3,0),(50,50,50)] NotLockable
uniform vec3 Ey_; slider[(-50,-50,-50),(0,0,0),(50,50,50)] NotLockable
uniform vec3 Target; slider[(-50,-50,-50),(0,0,0),(50,50,50)] NotLockable
uniform vec3 Targe_; slider[(-50,-50,-50),(0,0,0),(50,50,50)] NotLockable
uniform vec3 Up; slider[(0,0,0),(0,0,1),(0,0,0)] NotLockable

uniform int subframe;
uniform vec2 pixelSize;

varying vec3 from;
varying vec2 coord;
varying vec2 viewCoord;
varying vec2 viewCoord2;
varying vec3 dir;
varying vec3 Dir;
varying vec3 UpOrtho;
varying vec3 Right;
varying vec2 PixelScale;


#ifdef providesInit
void init(); // forward declare
#else
	void init() {}
#endif

vec2 rand(vec2 co){
	// implementation found at: lumina.sourceforge.net/Tutorials/Noise.html
	return
	vec2(fract(sin(dot(co.xy ,vec2(12.9898,78.233))) * 43758.5453),
	fract(cos(dot(co.xy ,vec2(4.898,7.23))) * 23421.631));
}

void main(void){
	
	// Prepare Camera
	float FOVP = FOV; // perspective FOV (0-2)
	if (UseTrueFOV) FOVP = tan(radians(TrueFOV/2));
	if (UseFocalLength) FOVP = FilmGate / (2 * FocalLength);
	
	
	gl_Position =  gl_Vertex;
	coord = (gl_ProjectionMatrix*gl_Vertex).xy;
	coord.x*= pixelSize.y/pixelSize.x;
	
	// we will only use gl_ProjectionMatrix to scale and translate, so the following should be OK.
	PixelScale = vec2(pixelSize.x*gl_ProjectionMatrix[0][0], pixelSize.y*gl_ProjectionMatrix[1][1]);
	viewCoord = gl_Vertex.xy;
	viewCoord2 = (gl_ProjectionMatrix*gl_Vertex).xy;
	
	vec3 Upn;
	
	#ifdef USE_EXTERNAL_CAMERA
		if (iCamWeight>0){
			frame = int(ceil(iCamFrame));
			from =  getCameraPos()* iCamWeight + Eye * (1-iCamWeight);
			Dir = getCameraFw() * iCamWeight + normalize(Target - Eye) * (1-iCamWeight);
			Upn = getCameraUp() * iCamWeight + normalize(Up) * (1-iCamWeight);
			UpOrtho = normalize(Upn-dot(Dir,Upn)*Dir);
			Right = getCameraRight();
		}
		#else
			from = Eye + Ey_;
			Dir = normalize((Target+Targe_) - (Eye+Ey_));
			Upn = normalize(Up);
			UpOrtho = normalize( Upn-dot(Dir,Upn)*Dir );
			Right = normalize(cross(Dir,UpOrtho));
			//}
		#endif
		
		coord *= FOVP;
		init();
		
}
#endvertex

#group Raytracer

#define PI  3.14159265358979323846264
// Camera position and target.
varying vec3 from;
varying vec3 dir;
varying vec3 dirDx;
varying vec3 dirDy;
varying vec2 coord;
varying float zoom;

// Field of view calculations
varying float FOVP; // perspective FOV (0-2)
varying float FOVT; // Horizontal FOV in degrees
varying float FOVFL; // FOV using filmgate and focal length

uniform int subframe;
uniform sampler2D backbuffer;
//uniform sampler2D tagBuffer;
//uniform sampler2D zbuffer;

varying vec2 viewCoord;
varying vec2 viewCoord2;
varying vec3 Dir;
varying vec3 UpOrtho;
varying vec3 Right;

//Distance from camera to in focus zone
uniform float FocalPlane; slider[0.01,1,50]
uniform float Aperture; slider[0,0.00,0.5]
//Width of the in focus zone. this value is relative to FocalPlane
uniform float InFocusAWidth; slider[0,0,1]
//Number of sides for the diaphragm. 2->circular
uniform int ApertureNbrSides; slider[2,2,10] Locked
//Rotation of the diaphragm
uniform float ApertureRot; slider[0,0,360]
//For star shaped diphragms. Very limited
uniform bool ApStarShaped; checkbox[false] Locked

vec2 rand2(vec2 co){
	// implementation found at: lumina.sourceforge.net/Tutorials/Noise.html
	return
	vec2(fract(sin(dot(co.xy ,vec2(12.9898,78.233))) * 43758.5453),
	fract(cos(dot(co.xy ,vec2(4.898,7.23))) * 23421.631));
}

vec2 NGonDisc(vec2 co) {
	vec2 r = rand2(co);
	float d=1./tan(PI/float(ApertureNbrSides));
	float lh=1./sqrt(1.+d*d);
	vec2 p=vec2(r.x+r.y, r.x-r.y);
	if(!ApStarShaped) p.x=1.-abs(p.x-1.);
	p.x*=d;
	p*=lh;
	float a=dot(rand2(11.*co+vec2(1.,1.)),vec2(1.,1.));
	a=floor(a*float(ApertureNbrSides))*2.*PI/float(ApertureNbrSides) + ApertureRot*PI/360.;
	vec2 cs= vec2(cos(a), sin(a));
	p= vec2(cs.x*p.x-cs.y*p.y, cs.x*p.y+cs.y*p.x);
	/*a=ApertureRot*PI/360.;
	cs= vec2(cos(a), sin(a));
	p= vec2(cs.x*p.x-cs.y*p.y, cs.x*p.y+cs.y*p.x);*/
	return p;
}

vec2 uniformDisc(vec2 co) {
	vec2 r = rand2(co);
	return sqrt(r.y)*vec2(cos(r.x*6.28),sin(r.x*6.28));
}

#ifdef providesInit
void init(); // forward declare
#else
	void init() {}
#endif
//out vec4 gl_FragColor;

struct SRay{
	vec3 Origin; // origin of the ray. Intially camera center
	vec3 Direction; // main ray direction
	vec3 Offset; // For DoF
	float Pos; // current position along the ray
	float fudge;// Lazy way to avoid overshooting
	float iFP;// =1/FocalPlane
};

struct SColor{
	vec3 RGB; // RGB
	float A; // Alpha
	float Z; // Depth
	vec3 Tag; // RGB Color used for tagging layers
	bool DoPost; // Whether it should postprocess the pixel
};

struct SCamera {
	vec3 Position;
	vec3 Direction;
	vec3 Target;
	float FocalLength;
	float FilmWidth;
	float FilmHeight;
};

// Given a camera pointing in 'dir' with an orthogonal 'up' and 'right' vector
// and a point, coord, in screen coordinates from (-1,-1) to (1,1),
// a ray tracer direction is returned

vec3 equiRectangularDirection(vec2 coord, vec3 dir, vec3 up, vec3 right)  {
	vec2 r = vec2(coord.x,(1.0-coord.y)*0.5)*PI;
	return cos(r.x)*sin(r.y)*dir+
	sin(r.x)*sin(r.y)*right+
	cos(r.y)*up;
}

#ifdef providesSRCurrentPt
vec3 SRCurrentPt(SRay Ray);
#else
	vec3 SRCurrentPt(SRay Ray){
		float t = Ray.Pos;
		return Ray.Origin+Ray.Direction*t + Ray.Offset*(1.-t*Ray.iFP + clamp(t*Ray.iFP-1., -InFocusAWidth, InFocusAWidth));
	}
#endif

void SRAdvance(inout SRay Ray, float dist){
	Ray.Pos+=dist*Ray.fudge;
}

SRay SRReflect(SRay Ray, vec3 normal, float eps){
	vec3 hit=SRCurrentPt(Ray);
	Ray.Direction = reflect(Ray.Direction, normal);
	Ray.Offset = reflect(Ray.Offset, normal);
	Ray.Origin= hit + reflect(Ray.Origin - hit, normal);//hit-Ray.Direction*Ray.Pos;
	Ray.Pos+=eps;
	return Ray;
}

// ------------------------------------------------- Post ---------------------------------
#group Post
uniform bool DryPass; checkbox[false]
uniform float Gamma; slider[0.0,1.0,5.0]
// 1: Linear, 2: Expontial, 3: Filmic, 4: Reinhart; 5: Syntopia
uniform int ToneMapping; slider[1,3,5]
uniform float Exposure; slider[0.0,1.0,3.0]
uniform float Brightness; slider[0.0,1.0,5.0];
uniform float Contrast; slider[0.0,1.0,5.0];
uniform float Saturation; slider[0.0,1.0,5.0];
uniform float GaussianWeight; slider[0.0,1.0,10.0];


uniform float AntiAliasScale; slider[0,2,10]
//Apply Bloom. Note: specularity parameters (Light tab) have a huge impact.
uniform bool 	Bloom; checkbox[false]
uniform float	BloomIntensity; slider[0,0.25,2]
uniform float	BloomPow; slider[0,2,10]
uniform int   	BloomTaps; slider[1,4,20]

varying vec2 PixelScale;

uniform float FOV;
uniform bool UseTrueFOV;
uniform float TrueFOV;
uniform bool UseFocalLength;
uniform float FocalLength;
uniform float FilmGate;

/*
uniform bool FlatPass;
uniform float Gamma;
uniform float Exposure;
uniform float Brightness;
uniform float Contrast;
uniform float Saturation;
uniform int ToneMapping;

varying vec2 coord;
uniform sampler2D frontbuffer;
*/

/*
** Based on: http://mouaif.wordpress.com/2009/01/22/photoshop-gamma-correction-shader/
**
** Contrast, saturation, brightness
** Code of this function is from TGM's shader pack
** http://irrlicht.sourceforge.net/phpBB2/viewtopic.php?t=21057
*/
// For all settings: 1.0 = 100% 0.5=50% 1.5 = 150%
vec3 ContrastSaturationBrightness(vec3  color, float brt, float sat, float con) {
	const vec3 LumCoeff = vec3(0.2125, 0.7154, 0.0721);
	vec3 AvgLumin = vec3(0.5);
	vec3 brtColor = color * brt;
	float intensityf = dot(brtColor, LumCoeff);
	vec3 intensity = vec3(intensityf, intensityf, intensityf);
	vec3 satColor = mix(intensity, brtColor, sat);
	vec3 conColor = mix(AvgLumin, satColor, con);
	return conColor;
}

float sigmoid(float t) {
	float K = 1.0-1./(1.0+exp(-0.5*Contrast*5.));
	t -= 0.5;
	float  x = 1./(1.0+exp(-t*Contrast*5.))-K;
	return x/((1.0-2.0*K));
}

vec3 sigmoid3(vec3 t) {
	return vec3(sigmoid(t.x),sigmoid(t.y),sigmoid(t.z));
}

vec3 applyPost(vec3 c){
	if (ToneMapping==1) {
		// Linear
		c = c*Exposure;
		c = ContrastSaturationBrightness(c, Brightness, Saturation, Contrast);
	
	} else if (ToneMapping==2) {
		// ExponentialExposure
		c = vec3(1.0)-exp(-c*Exposure);
		c = ContrastSaturationBrightness(c, Brightness, Saturation, Contrast);
	
	} else if (ToneMapping==3) {
		// Filmic: http://filmicgames.com/archives/75
		c*=Exposure;
		vec3 x = max(vec3(0.),c-vec3(0.004));
		c = (x*(6.2*x+.5))/(x*(6.2*x+1.7)+0.06);
		c = pow(c, vec3(2.2)); // It already takes the Gamma into acount
		c = ContrastSaturationBrightness(c, Brightness, Saturation, Contrast);
	
	} else if (ToneMapping==4) {
		// Reinhart
		c*=Exposure;
		c = c/(1.+c);
	c = ContrastSaturationBrightness(c, Brightness, Saturation, Contrast);
	}  else if (ToneMapping==5) {
		c = sigmoid3(c*Exposure+vec3(Brightness-1.0));
		
	}
	c = pow(c, vec3(1.0/Gamma));
	return c;
	// return  vec3(1.,0,0); 
}

/// ------------------------------------------------- Main ---------------------------------

// implement this;
// vec3 color(SRay Ray);
SColor color(SRay Ray);

void main() {
	
	init();
	
	// Prepare Camera
	float FOVP = FOV; // perspective FOV (0-2)
	if (UseTrueFOV) FOVP = tan(radians(TrueFOV/2));
	if (UseFocalLength) FOVP = FilmGate / (2 * FocalLength);
	vec3 hitNormal = vec3(0.0);
	vec3 hit;
	
	// A Thin Lens camera model with Depth-Of-Field
	
	vec2 r = vec2(0.);
	if(ApertureNbrSides == 2)
	// We want to sample a circular diaphragm
	r = Aperture*uniformDisc(viewCoord*(float(subframe)+1.0));
	else
	// We want to sample a polygonal diaphragm
	r = Aperture*NGonDisc(viewCoord*(float(subframe)+1.0));
	
	//r*=1.-InFocusAWidth;//an error while modifying this script. It gives interresting result though.
	
	//jitter for multisampling
	vec2 disc = uniformDisc(coord*float(1+subframe)); // subsample jitter
	vec2 jitteredCoord = coord + AntiAliasScale*PixelScale.y*FOVP*disc;
	
	// Offset for DoF
	vec3 lensOffset =  r.x*Right + r.y*UpOrtho;
	
	//"principal" ray direction
	vec3 rayDir = (Dir+ jitteredCoord.x * Right + jitteredCoord.y * UpOrtho);
	float rayDirLength=length(rayDir);
	float FPO1= length( rayDir+lensOffset*( 1. - 1./FocalPlane + clamp( 1./FocalPlane-1., -InFocusAWidth, InFocusAWidth)));
	
	//Construct the ray
	SRay Ray = SRay(from, rayDir/rayDirLength, lensOffset, 0., 1./max(1.,FPO1), 1./FocalPlane);
	
	//grab incoming light
	SColor c =  color(Ray);
	
	// Check for contents and do postprocessing if applicable
	
	if (!DryPass && c.DoPost){c.RGB = applyPost(c.RGB);}
	
	// Accumulate
	vec4 prev = texture2D(backbuffer,(viewCoord+vec2(1.0))/2.0);
	
	float w =1.0-length(disc);
	if (GaussianWeight>0.) {
		w = exp(-dot(disc,disc)/GaussianWeight)-exp(-1./GaussianWeight);
	}
	
	// Sets the final color
	gl_FragColor = prev + vec4(c.RGB*w, w);
}
