#donotrun
#info DE-Raytracer-PK-1.0.9.frag by Patryk Kizny
#info Original shader by syntopia. Modifications by knighty + Eiffie + ChrisJRodgers + Patryk Kizny
#info  - Added multiple reflections, multilayer clouds and directional sunlight + point spotlight
#info  - Added color tagging of layers & depth pass.
#info Accepts the following define statements:
#info - TAG_LAYERS - Enables layer object tagging by color
#info - DEPTH_PASS - Enables extra pass to calculate depth
#info - USE_KN_VOLUMETRIC - Enables volumentric behaviour of spotlight
#info - USE_IQ_CLOUDS - Enables Inigo Quilez Clouds
#info - USE_IQ_CLOUDS_MULTILAYER - Enables MultiLayer Clouds 
#info - USE_EIFFIE_SHADOW - slightly faster, but banding issues occurs
#info - USE_KN2_SHADER - makes use of older KN2 shader, else uses PK improved shading (slower)
#info - USE_PK_BACKGROUND - adds flat or equirect background support
#info - USE_PK_IBL - adds image-based lighting
#info - providesColor, implement vec3 baseColor(hit,hitNormal); // RGB
#info - providesBackground, implement vec4 backgroundColor(vec3 Ray.Dir, vec2 pos); // RGBA
#info - providesFloor, implement vec4 floorColor(vec3 pos); // RGBA
#info - providesNormal, implement vec3 normal(vec3 pos, float normalDistance);

// ---- TODOs & Bugs
// 2015/07/26
// - Reflections are fucked after adding SColor
// - Add matte objects to receive shadows
// - Add image based lighting
// - Testing of fog & light is probably required
// - Light Glow does not blend properly against background

// - Move backgroundcolor to separate function. Optimize
// --- move bitmaps to background tab, add switch, add flat/equirect mapping

// ---- Log Patryk Kizny Changes -----

// 2015/07/24
// - Modified vec3 color to struct SColor
// - Moved Depth and Tagging out of Post processing
// - Requires use of 3DPK 1.0.4 or later

// 2015/07/24
// - Modified vec3 color to struct SColor
// - Moved Depth and Tagging out of Post processing
// - Requires use of 3DPK 1.0.4 or later

// 2015/07/22
// - Adding background

// 2015/07/15
// - Aded depth pass

// 2015/07/14
// - Aded color tagging of layers

// 2015/07/12
// - Fixed AO color - now coloredAO color is included for spot and sun

// 2015/06/29
// - Specular is now dependent on the intensity of sunlight and spotlight

// 2015/06/23
// - Added more intuitive camera FOV

// 2015/06/20
// - added multilayer clouds
// - added back directional spotlight (sunlight) working together with point spotlight
// - shadows seem to work fine together
// - Todo: shadows can be probably computed faster - did not optimize it

// ---- Log -----

//mar 7 2015:
// - Now the fog is colored and effects lighting.
// - Colored reflections.
//mar 6 2015:
// - Black dots again. Visible especially when using reflections. Got hard time to fix it. :-/. Any time a transcendental function is involved, one have to take care about infinities, denormals and NaNs.
//mar 5 2015:
// - Implemented stratified sampling for volumetric light and "corrected" use of RNG.
//mar 3 2015:
// -Added Iq's cloud implemented by Eifie. In order to use it #define USE_IQ_CLOUDS before including DE-kn2.frag
// -Added volumetric light. In order to use it #define USE_KN_VOLUMETRIC before including DE-kn2.frag.
//mar 1 2015:
// -Updated settings.
// -Tweaked the AO function to make it "multiscale". Work still in progress: Identify and implement pertinent parameteres. See Baird-Delta preset "AO_Alone"
// -Corrected warnings.
//feb 24 2015:
// -Corrected some bugs. Floating point numbers are tricky.
// -Corrected bug with DoF and Floor.
// -Added point kight distance falloff.
// -Modified height fog formula again.
//feb 23 2015:
// -switched to Eiffie's shoft shadow. Looks much better to me.
// -added altitude fog.
// -ChristRodgers added Point light with glow. Modified a bit.
// -Added bloom effect. Bug corrected.
//Known bugs:
// -Still some black spots on some graphics card. Maybe a driver issue. Don't use unrealistic values for DoF effect plz ;o)
//ToDo:
// -Simplify point light glow computations. Low priority: The light glow is good but suffers from banding artifacts (needs high dither values).
// -Implement user defined matrials. High priority.
// -Reduce banding artifacts ue to tracing accuracy. High priority.
// -SoC? Global illumination? (maybe needs complete refactoring and/or removing InFocusAWidth which may complicate things)

#include "3D-PK-1.0.5.frag"

#group Raytracer
// Distance to object at which raymarching stops.
uniform float Detail;slider[-7,-3.5,0];
// The step size when sampling AO (set to 0 for old AO)
uniform float DetailAO;slider[-7,-0.5,0];
// Sets focal plane to Target location
uniform bool AutoFocus; checkbox[false]
// Lower this if the system is missing details
uniform float FudgeFactor;slider[0,0.95,1];
// Maximum number of raymarching steps.
uniform int MaxRaySteps; slider[0,256,5000]
// Maximum distance at which raymarching stops
uniform float MaxDistance; slider[0,32,1000];
// Use this to boost Ambient Occlusion and Glow
uniform float  MaxRayStepsDiv;  slider[0,1.8,10]
// Can be used to remove banding
uniform float Dither;slider[0,0.5,1];
// Used to prevent normals from being evaluated inside objects.
uniform float NormalBackStep; slider[0,1,10]

float minDist = pow(10.0,Detail);
float aoEps = pow(10.0,DetailAO);
vec4 orbitTrap = vec4(10000.0);
//float fractionalCount = 0.0;

#ifdef DEPTH_PASS
	#group Depth
	uniform float DepthScale; slider[-1.0,0.0,1.0]
	uniform float DepthOffset; slider[-5.0,0.0,5.0]
	uniform float DepthOffsetPost; slider[-1.0,0.0,1.0]
	uniform int DepthScalePower; slider[0,2,16]
	uniform bool  MultiChannelDepth; checkbox[false]
	uniform bool FlatPass; checkbox[false]
	uniform bool OutputDepth; checkbox[false]
#endif

#ifdef TAG_LAYERS
	#group Layer Tag
	uniform vec3 tagBackground; color[0.0,0.0,0.0]
	uniform vec3 tagLight; color[1.0,0.0,0.0]
	uniform vec3 tagFloor; color[0.0,1.0,0.0]
	uniform vec3 tagObject; color[0.0,1.0,1.0]
	uniform vec3 tagFractal; color[1.0,1.0,1.0]
	uniform bool OutputTag; checkbox[false]
#endif

#group Material Floor
// Diffuse material
uniform vec3 fDiffuse; color[0.5,0.5,0.5]
// Diffuse falloff
//uniform float fRoughness; slider[0,0.1,1];
// Diffuse falloff
uniform float fDiffExp; slider[0,1,4];
// Specular power
uniform vec4 fSpecular; color[0,0.5,1,1,1,1];
// Specularity roughness
uniform float fSpecularExp; slider[1,16,256];
// Max specular
uniform float fSpecularMax; slider[0,2,8];

#group Material Object
// Diffuse material
uniform vec3 oDiffuse; color[0.5,0.5,0.5]
// Diffuse falloff
//uniform float oRoughness; slider[0,0.1,1];
// Diffuse falloff
uniform float oDiffExp; slider[0,1,4];
// Specular power
uniform vec4 oSpecular; color[0.0,0.5,1.0,1.0,1.0,1.0];
// Specularity roughness
uniform float oSpecularExp; slider[1,16,256];
// Max specular
uniform float oSpecularMax; slider[0,2,8];

struct Material{
	vec3 diffuse;
	vec4 specular;
	float diffExp;
	float specExp;
	float specMax;
	// float roughness;
	// vec2 anisotropy;
};


Material ObjMat = Material(oDiffuse, oSpecular, oDiffExp, oSpecularExp, oSpecularMax);
Material FloorMat = Material(fDiffuse, fSpecular, fDiffExp, fSpecularExp, fSpecularMax);

#group Light
// AO based on the number of raymarching steps
uniform vec4 AO; color[0,0.7,1,0.0,0.0,0.0];
// Correction of AO
uniform float AoCorrect; slider[0.0,0.0,1.0]
//  Color and power of the sunlight
uniform vec4 SunLight; color[0.0,0.4,1.0,1.0,1.0,1.0];
// Direction to the sunlight (spherical coordinates)
uniform vec2 SunDir;  slider[(-1,-1),(0.1,0.1),(1,1)];
// Use this to show position of the sun
uniform bool DebugSun; checkbox[false] Locked
// Quality shdows for sunlight only
// uniform bool QualityShadows; checkbox[true];
// Light coming from the camera position (diffuse lightning)
uniform vec4 AmbientLight; color[0,1,2,1.0,1.0,1.0];
// Controls the minimum ambient light, regardless of directionality
uniform float AmbientLightMin; slider[0.0,0.0,1.0]
// Glow is not used anymore //REMOVE
// Glow based on distance from fractal
uniform vec4 Glow; color[0,0.0,1,1.0,1.0,1.0];
//uniform vec4 InnerGlow; color[0,0.0,1,1.0,1.0,1.0];
uniform int GlowMax; slider[0,20,1000]
// Adds old-type fog based on distance
uniform float Fog; slider[0,0.0,2]
uniform float FogExp; slider[0,2,10]
// Value of reflections, black = no reflections
uniform vec3 Reflection; color[1,1,1]
// Number of reflections. Slows down rendering
uniform int ReflectionsNumber; slider[0,0,5]
// Hard shadows shape is controlled by SpotDir + SunLightDir
uniform float HardShadow; slider[0,0,1]
// Higher values gove harder shadows
uniform float ShadowSoft; slider[0.0,10.0,20]

#group Spot Light
// Stpolight glow
uniform bool SpotGlow; checkbox[true]
// Color and strength of the spot light
uniform vec4 Spot; color[0.0,1.,10.0,1.0,1.0,1.0];
// 3D Position of the spot light
uniform vec3 SpotPos; slider[(-10,-10,-10),(0,0,0),(10,10,10)]
// Size of Spotlight
uniform float SpotSize; slider[0.0,0.1,1]
// Falloff
uniform float SpotFallOff; slider[0,0,2]
// Glow radius
uniform float SpotGlowRad; slider[0.0,0.0,5.0]
// Glow exponent
uniform float SpotGlowExp; slider[0.0,1.0,5.0]

// define few vars to trigger shadow computation
bool UseSpot = bool(Spot.w>0);
bool UseSun =  bool(SunLight.w>0);
bool CastShadows = bool((Spot.w + SunLight.w)>0 && HardShadow > 0 );
float sunShadowStrength;

#group Coloring
// This is the pure color of object (in white light)
uniform vec3 BaseColor; color[1.0,1.0,1.0];
// Determines the mix between pure light coloring and pure orbit trap coloring
uniform float OrbitStrength; slider[0,0,1]
// Closest distance to YZ-plane during orbit
uniform vec4 X; color[-1,0.7,1,0.5,0.6,0.6];
// Closest distance to XZ-plane during orbit
uniform vec4 Y; color[-1,0.4,1,1.0,0.6,0.0];
// Closest distance to XY-plane during orbit
uniform vec4 Z; color[-1,0.5,1,0.8,0.78,1.0];
// Closest distance to  origin during orbit
uniform vec4 R; color[-1,0.12,1,0.4,0.7,1.0];
uniform bool CycleColors; checkbox[false]
uniform float Cycles; slider[0.1,1.1,32.3]

#group Floor
uniform bool FloorEnable; checkbox[false]
// Makes floor transparent and take shadows
uniform bool FloorMatte; checkbox[false]
uniform bool FloorDebug; checkbox[false]
uniform vec3 FloorNormal; slider[(-1,-1,-1),(0,0,1),(1,1,1)]
uniform float FloorHeight; slider[-5,0,5]
uniform vec4 FloorColor; color[0.,1.,1.,1.,1.,1.]

bool floorHit = false;
float floorDist = 0.0;
vec3 floorNormal = normalize(FloorNormal);
float fSteps = 0.0;

#group Height_Fog
uniform float HF_Fallof; slider[0.0005,0.1,5.]
uniform float HF_Const; slider[0.0,0.0,1.0]
uniform float HF_Intensity; slider[0.0,0.0,1.0]
uniform vec3 HF_Dir; slider[(-1,-1,-1),(0,0,1),(1,1,1)]
uniform float HF_Offset; slider[-10.,0.0,10.0]
uniform vec4 HF_Color; color[0.0,1.,3.0,1.0,1.0,1.0]

#ifdef USE_KN_VOLUMETRIC
	//to modify volumetric light intensity
	uniform float HF_Scatter; slider[0.0,0.,10.0]
	//Anisotropy of the fog.
	//uniform float HF_Anisotropy; slider[0.0,0.,1.0]
	uniform vec3 HF_Anisotropy; color[0,0,0]
	//for stratified sampling. Helps to accelerate convergence but slows down the framerate. Better use it for final render. You'll need less subframes.
	uniform int HF_FogIter; slider[1,1,16]
	//Is shadow cast for every sample. Slows down the rendering.
	uniform bool HF_CastShadow; checkbox[false]
#endif

#ifdef USE_PK_BACKGROUND
	#group Background
	// Background color
	uniform vec4 BgColor; color[0.,1.,1.,0.5,0.5,0.5]
	// Vignette background
	uniform float BgGradient; slider[0.0,0.3,5.0]
	// 
	uniform bool BgUseMap; checkbox[false]
	uniform bool BgMapEquirectangular; checkbox[false]
	
	uniform sampler2D BgMap; file[Ditch-River_2k.hdr]
	uniform vec3 BgMapScale; slider[(0,0,0),(1,1,1),(1,1,1)]
uniform vec3 BgMapOffset; slider[(0,0,0),(0,0,0),(1,1,1)]
#else
	#group Coloring
	uniform vec4 BgColor; color[0.,1.,1.,0.5,0.5,0.5]
#endif

#ifdef USE_PK_IBL
	#group Environment Lighting
	// If cheked will use backgroundColor to sample environment. If using map it must be equirectangular.
	uniform bool EnvUseBackground; checkbox[true]
	// Equirectangular map for environment specular mapping
	uniform sampler2D EnvMapSpecular; file[Ditch-River_2k.hdr]
	// Equirectangular map for environment diffuse mapping - use a blurred map for efficiency
	uniform sampler2D EnvMapDiffuse; file[Ditch-River_Env.hdr]
	// Enables use of alternate map for diffussion channel
	uniform bool EnvUseDiffusionMap; checkbox[true]
	// Dampens diffusion environment light in specific directions
	uniform vec3 EnvLightFactor; slider[(0,0,0),(1,1,1),(1,1,1)]
	// fades IBL in diffussion channel
	uniform float EnvDiffussionFade; slider[0.,1.,1.]
	// fades IBL in specular channel
	uniform float EnvSpecularFade; slider[0.,0.1,1.]

	// Todo
	//uniform vec3 EnvMapScale; slider[(0,0,0),(1,1,1),(1,1,1)]
	//uniform vec3 EnvMapOffset; slider[(0,0,0),(1,1,1),(1,1,1)]
	
#endif


// ------------------------------------------------------ Background -----------------------------------------

// Convert dir to spherical coordinates.
vec3 equirectangularMap(sampler2D sampler, vec3 dir) {
	dir = normalize(dir);
	vec2 longlat = vec2(atan(dir.y,dir.x),acos(dir.z));
	// Normalize, and lookup in equirectangular map.
	return texture2D(sampler,longlat/vec2(2.0*PI,PI)).xyz;
}

#ifdef USE_PK_BACKGROUND
	vec4 backgroundColor(vec3 dir, vec2 pos) {
		vec4 bg = vec4(0.);
		#ifdef providesBackground
		bg = backgroundColor(dir, pos);
		#else 
			bg = BgColor;
			// Todo perhaps can include transparency
			if (BgUseMap){
				if (BgMapEquirectangular){
				bg.xyz = equirectangularMap(BgMap,dir * BgMapScale + BgMapOffset);
				} else {
					bg.xyz = texture2D(BgMap,pos * BgMapScale.xy + BgMapOffset.xy).xyz;
				}
				bg.w = 1.0;	
			}
		#endif
		
		if (BgGradient > 0.0) {
			bg.xyz = mix(bg.xyz, vec3(0.0,0.0,0.0), length(coord)*BgGradient);
		}
		return bg;
	}
	#else
		vec4 backgroundColor(vec3 dir, vec2 pos) {
			return BgColor;
		}
	#endif
	
	
	// ----------------------------------------- DE Section --------------------------------------------
	
	float DE(vec3 pos) ; // Must be implemented in other file
	
	
	float DElight(vec3 pos, vec3 lightPos) {
		return length(lightPos-pos)-SpotSize;
	}
	
	float DEL(vec3 pos, inout float lightde, vec3 lightPos) {
		
		lightde=DElight(pos, lightPos)-SpotSize;
		return min(DE(pos)*FudgeFactor,lightde);
		
	}
	
	float DEL2(vec3 pos, inout float lightde, vec3 lightPos) {
		
		lightde=DElight(pos, lightPos)-SpotSize;
		return min(DE(pos),lightde);
	}
	
	
	float DEF(vec3 p) {
		
		float d = DE(p) ;
		if (FloorEnable) {
			floorDist = abs(dot(floorNormal,p) - FloorHeight);
			if (d<floorDist) {
				fSteps++;
				return d;
			}
			
			else return floorDist;
		}
		
		else {
			fSteps++;
			return d;
		}
	}
	
	float DEF2(vec3 p) {
		
		if (FloorEnable) {
			floorDist = abs(dot(floorNormal,p)-FloorHeight);
			return min(floorDist, DE(p));
		}
		
		else {
			return DE(p);
		}
		
	}
	
	
	#ifdef providesNormal
	vec3 normal(vec3 pos, float normalDistance);
	#else
		vec3 normal(vec3 pos, float normalDistance) {
			normalDistance = max(normalDistance*0.5, 1.0e-5);
			vec3 e = vec3(0.0,normalDistance,0.);
			vec3 n = vec3(DE(pos+e.yxx)-DE(pos-e.yxx),
			DE(pos+e.xyx)-DE(pos-e.xyx),
			DE(pos+e.xxy)-DE(pos-e.xxy));
			n = normalize(n);
			return n==n ? n : vec3(0.0);
		}
		
	#endif
	
	#ifdef providesBackground
		vec4 backgroundColor(vec3 dir, vec2 pos);
	#endif
	
	float exp1(float x){ return exp(clamp(x,-80.,80.));}
	
	//have to use this to avoid black dots (again)
	vec3 exp1(vec3 x){ return exp(clamp(x,vec3(-80.),vec3(80.)));}
	
	float fogAmount(vec3 P0, vec3 P1){//Modified from: http://www.iquilezles.org/www/articles/fog/fog.htm
		//ToDo: make dependent on the fog color
		vec3 hfdir=normalize(HF_Dir);
		float t=length(P1-P0);
		float A=HF_Fallof*dot(P1-P0,hfdir);
		A=(1.0-exp1(-A))/A;
		float amount= (HF_Intensity * exp1(-HF_Fallof*(dot(P0,hfdir)-HF_Offset))* A  + HF_Const) * t;
		//amount=clamp(amount,-5.,100.);
		return clamp(exp1(-amount),0.,1.);//Return transmission factor
	}
	
	vec3 fogAmount3(vec3 P0, vec3 P1){//Modified from: http://www.iquilezles.org/www/articles/fog/fog.htm
		
		//ToDo: make dependent on the fog color
		vec3 hfdir=normalize(HF_Dir);
		float t=length(P1-P0);
		float A=HF_Fallof*dot(P1-P0,hfdir);
		A=(1.0-exp1(-A))/A;
		vec3 amount= HF_Color.rgb*(HF_Intensity * exp1(-HF_Fallof*(dot(P0,hfdir)-HF_Offset))* A  + HF_Const) * t;
		//amount=clamp(amount,-5.,100.);
		return clamp(exp1(-amount),vec3(0.),vec3(1.));//Return transmission factor
		
	}
	
	vec3 ptLightGlow(float glow,  vec4 lightColor){
		
		if(SpotGlow){
			float glow1=exp(-pow(SpotGlowRad+0.0001,-2.)*pow(glow,SpotGlowExp));
			float glow2=exp(-20.*pow(glow,1.));
			return lightColor.xyz*glow2*lightColor.w*1.+ lightColor.xyz*glow1;
		}
		
		return vec3(0.);
		
	}
	
	#define MIN_EPS 2./16777216.
	
	float rand(vec2 co){
		// implementation found at: lumina.sourceforge.net/Tutorials/Noise.html
		return fract(sin(dot(co.xy ,vec2(12.9898,78.233))) * 43758.5453);
	}
	
	#ifdef USE_EIFFIE_SHADOW
		// Uses the soft-shadow approach by Eiffie:
		// Produces banding artifacts for close spotlight // todo
		
		float linstep(float a, float b, float t){return clamp((t-a)/(b-a),0.,1.);}
		
		float shadow(vec3 ro, vec3 lightPos, float eps){//float FuzzyShadow(vec3 ro, vec3 rd, float lightDist, float coneGrad, float rCoC){
			//return 1.;
			float rCoC=eps;
			vec3 rd=(lightPos-ro);
			float lightDist=length(rd);
			rd/=lightDist;
			float coneGrad=ShadowSoft;
			if(ShadowSoft==0.) coneGrad=SpotSize/lightDist; else coneGrad=1./ShadowSoft;
			float t=DEF2(ro)+rCoC;//avoid self shadowing
			float d=1.0,s=1.0;
			float jitter=Dither*(rand(ro.xy*(float(subframe)+1.0))-0.5);
			for(int i=0;i<MaxRaySteps;i++){
				if(t>lightDist || s<0.001) break;
				float r=rCoC+t*coneGrad;//radius of cone
				#ifndef PERFECT_DE
					//Because most of the time (in particular with fractals) DE is not well defined inside. The soft shadow doesn't "bleed" inside.
					d=DEF2(ro+rd*(t+r*jitter));
					s*=linstep(-r,r,d);//smoothstep(-r,r,d);//
				t+=abs(0.75*d+.25*r);//*mix(1.,0.2*rand(gl_FragCoord.xy*0.02*vec2(i)*t),jitter);
				#else
					//in the case the DE gives correct interior distance this is the "correct" formula giving "correct" shadow.
					d=DEF2(ro+rd*(t+r*jitter))+r;
					s*=linstep(0.,2.*r,1.*d);//smoothstep(-r,r,d);//
					t+=abs(0.75*d+.25*r);
				#endif
			}
			
			s=max(0.,s-0.001)/0.999;
			return clamp(1.-s,0.0,1.0);
		}
	
	#else
		// Uses the soft-shadow approach by Quilez:
		// http://iquilezles.org/www/articles/rmshadows/rmshadows.htm
		float shadow(vec3 ro, vec3 lightPos, float eps) {
			vec3 rd=( lightPos-ro);
			float lightDist=length(rd);
			rd/=lightDist;
			float coneGrad=ShadowSoft;
			if(ShadowSoft==0.) coneGrad=lightDist/(SpotSize+0.001); else coneGrad=ShadowSoft;
			float totalDist =2.0*eps;
			float s = 1.0; // where 1.0 means no shadow!
			float jitter = Dither*(rand(ro.xy)-0.5);
			for (int steps=0; steps<MaxRaySteps && totalDist<lightDist; steps++) {
				vec3 p = ro + totalDist*(1.+jitter/coneGrad) * rd;
				float dist = DEF2(p);
				if (dist < eps)  return 1.0;
				s = min(s, coneGrad*(dist/totalDist));
				totalDist += dist;
			}
			
			return 1.0-s;
		}
		
	#endif
	
	// Old type shdows for the sunlight
	// Uses the soft-shadow approach by Quilez:
	// http://iquilezles.org/www/articles/rmshadows/rmshadows.htm
	float sunShadow(vec3 pos, vec3 sdir, float eps) {
		float totalDist =2.0*eps;
		float s = 1.0; // where 1.0 means no shadow!
		for (int steps=0; steps<MaxRaySteps && totalDist<MaxDistance; steps++) {
			vec3 p = pos + totalDist*sdir;
			float dist = DEF2(p);
			if (dist < eps)  return 1.0;
			s = min(s, ShadowSoft*pow((dist/totalDist),1.));
			totalDist += dist;
		}
		
		//return max(1.0-s,0);
		return 1.0-s;
	}
	
	// Ambient occlusion approximation.
	// Sample proximity at a few points in the direction of the normal.
	// tweaked to make it "multiscale".
	float ambientOcclusion(vec3 p, vec3 n) {
		float ao = 0.0;
		float de = DEF(p)/aoEps;//*(-DetailAO);//dividing by aoEps gives control on the overall scale of the AO.
		float wSum = 0.0;
		float w = 1.0;
		float d = 1.0-(Dither*rand(p.xy));
		float D=1.;
		for (float i =1.0; i <15.; i++) {
			// D is the distance estimate difference.
			// If we move 'n' units in the normal direction,
			// we would expect the DE difference to be 'n' larger -
			// unless there is some obstructing geometry in place
			float prevD=D;
			D = (DEF(p+ d*n*de) -0.*de)/(d*de);
			D=min(prevD,D);//if light is obscured at previous level it should be obscured at least at the same amount at this level.
			w *= 1.;//Maybe removed next time
			d*=1.5;//next level  -> x2. maybe becomes a parameter
			ao += w*clamp(1.0-D,0.0,1.0);
			wSum += w;
		}
		
		return clamp(AO.w*ao/wSum, 0.0, 1.0);
	}
	
	
	#ifdef USE_KN2_SHADER
		
		vec3 BRDF (Material m, vec3 n, vec3 v, vec3 l, vec4 light, vec3 coloredAO, float shadow){
			
			//  An attempt at Physcical Based Specular Shading:
			//  http://renderwonk.com/publications/s2010-shading-course/
			// (Blinn-Phong with Schickl term and physical normalization)
			
			vec3 h = normalize(l - v); // half vector
			float nDotL = max(0.0,dot(n, l));
			float nDotV = dot(n, v); //max(0.0,dot(n, v));
			float nDotH = max(0.0,dot(n, h));
			float ldotH = max(0.0,dot(l, h));
			
			// Fresnel functions
			// Todo should be modified for complex number maths! 
			float f0 = (m.specExp-1)/(m.specExp+1); f0*=f0;
			float fspec = f0 + (1-f0) * pow((1-nDotV),5);
			
			float diffuse = nDotL;
			float specular = ((m.specExp + 2.)/8.) * fspec * nDotL * pow(nDotH, m.specExp+0.00001) * m.specular.w;
			
			// Shadows & Spec limit
			if (CastShadows){
				diffuse = mix(diffuse, 0.0, HardShadow*shadow);
				specular = mix(specular, 0.0, HardShadow*shadow) * (1.-shadow); // diffuse and specular are different since we have soft lights
			}
			specular = min(m.specMax,specular);
			
			vec3 BRDF = (m.diffuse * diffuse * coloredAO + m.specular.xyz * specular) * light.xyz * light.w;	
			return BRDF;
			
		}
	
	#else
		
		vec3 BRDF (Material m, vec3 n, vec3 v, vec3 l, vec4 light, vec3 coloredAO, float shadow){
			
			// Bidirectional Reflectance Distribution Function
			// An attempt at Physcical Based Specular Shading:
			// http://renderwonk.com/publications/s2010-shading-course/
			// (Blinn-Phong with Schickl term and physical normalization)
			// Updated by Patryk Kizny
			
			/* Optimized BRDF Model 
			BRDF = diffCol/PI * (1 - Fdiff(f0)) + NN *  (Fspec(f0)* nDotH^n) / (max (nDotL, nDotV)
			fdiff(f0) = f0 + (1-f0)*(1-nDotV)^5
			fspec(f0) = f0 + (1-f0)*(1-nDotH)^5
			NN = (n+2) / 4* PI (2 - 2^(-n/2)) // Specular Standard normalization
			NN = 0.0397436 * n + 0.0856832 // 	Specular Optimized normalization
			1 - Fdiff(f0)) ==> (1-f0) // Diffuse Optimization
			*/
			m.specExp = max(m.specExp, 1.0);
			
			vec3 h = normalize(l - v); // half vector
			float nDotL = dot(n, l); // max(0.0,dot(n, l));
			float nDotV = dot(n, v); // max(0.0,dot(n, l));
			float nDotH = max(0.0,dot(n, h));
			float ldotH = max(0.0,dot(l, h));
			
			// Fresnel functions
			// f0 = ((1-n)/(1+n)) ^ 2 // n is the complex number refractive index
			// Todo should be modified for complex unmber maths! 
			float f0d = (m.diffExp-1)/(m.diffExp+1); f0d*=f0d; // v2
			float f0s = (m.specExp-1)/(m.specExp+1); f0s*=f0s; // v2
			float fdiff = f0d + (1-f0d) * pow((1-nDotL),5);
			float fspec = f0s + (1-f0s) * pow((1-nDotH),5);
			
			float NN =  0.0397436 * m.specExp + 0.0856832;
			
			vec3 diffuse = m.diffuse * coloredAO / PI * (1 - fdiff); // Use (1-f0) for optimization
			vec3 specular = m.specular.xyz * NN * fspec * pow(nDotH,m.specExp) / max(nDotL, nDotV) * m.specular.w;  // PK
			
			// Shadows
			
			if (CastShadows){
				diffuse = mix(diffuse, vec3(0.0), HardShadow*shadow);
				specular = mix(specular, vec3(0.0), HardShadow*shadow); // diffuse and specular are different since we have soft lights
				specular *=  1.-shadow;
			}
			specular = min(vec3(m.specMax),specular);
			
			vec3 BRDF = (diffuse + specular) *light.xyz *light.w;	
			return BRDF;
			
		}
	#endif
	
	#ifdef USE_PK_IBL
		vec3 BRDF_IBL (Material m, vec3 n, vec3 v, vec3 coloredAO){
			
			// Todo implement average sampling in multiple directions, taking into account roughness
			/// temporarily replaced by blurred map
			
			m.specExp = max(m.specExp, 1.0);
			
			vec3 rf = -2.0 * dot(v,n) * n + v; // reflected light
			vec3 h2 = normalize(rf - v); // half vector for reflected
			float nDotV = max(0.0,dot(n, v));
			float nDotH = max(0.0,dot(n, h2));
			
			float NN =  0.0397436 * m.specExp + 0.0856832;
			float f0 = (1-m.specExp)/(1+m.specExp); f0*=f0;
			float fspec = f0 + (1-f0) * pow((1-nDotV),5);
			
			vec3 diff = m.diffuse.xyz * coloredAO;
			vec3 spec = m.specular.xyz * m.specular.w * NN * (fspec * pow(nDotH,m.specExp)) / max(nDotH, nDotV);
			spec = min(spec, m.specMax);
			
			// If floor is enabled do a simple direction reflection
			
			vec3 sdn = n; // sampling direction along normal
			vec3 sdr = rf; // sampling direction along reflection

			if (FloorEnable){
				if (dot(n,FloorNormal)>0) sdn = 2 * dot(n,FloorNormal) * FloorNormal + n;
				if (dot(rf,FloorNormal)>0) sdr = 2 * dot(rf,FloorNormal) * FloorNormal + rf; // rf is negated above
			}
			
			vec3 dc = EnvUseDiffusionMap ? equirectangularMap(EnvMapDiffuse,sdn) : vec3(1);
			vec3 sc = EnvUseBackground ? backgroundColor(sdr, vec2(0,0)).xyz : equirectangularMap(EnvMapSpecular,sdr);
			
			dc = mix(vec3(0.), dc, dot(normalize(abs(sdn)), EnvLightFactor));
			//sc = mix(vec3(0.), sc, dot(sdr, EnvLightFactor));
			
			vec3 BRDF = diff * dc * EnvDiffussionFade + spec * sc * EnvSpecularFade;
			return BRDF;
			
			
		}
	#endif
	
	vec3 lighting(vec3 n, Material m, vec3 pos, vec3 dir, float eps, out float shadowStrength, float ao) {
		
		// ------------------- Base section
		
		// Ambient light (outside of BRDF)
		vec3 coloredAO = 1.0 - ao*(1-AO.xyz);
		float amb = max(AmbientLightMin, dot(-n, dir) * AmbientLight.w);
		vec3 ambient = AmbientLight.xyz * amb * m.diffuse * coloredAO * AmbientLight.w;
		
		// Sunlight
		vec3 sunDir = normalize(vec3(sin(SunDir.x*PI)*cos(SunDir.y*PI/2.0), sin(SunDir.y*PI/2.0)*sin(SunDir.x*PI), cos(SunDir.x*PI)));
		float sunShad = sunShadow(pos+n*eps, sunDir, eps);
		vec3 sunBRDF = BRDF(m,n,dir,sunDir,SunLight,coloredAO,sunShad);
		
		// Spotlight
		vec3 spotDir = normalize(SpotPos-pos);
		float spotShad = shadow(pos+n*eps, SpotPos, eps);
		vec3 spotBRDF = BRDF(m,n,dir,spotDir,Spot,coloredAO, spotShad);
		
		float D2L2 = dot(SpotPos-pos, SpotPos-pos);
		float spotFalloff = pow(D2L2,-SpotFallOff);
		vec3 FG = fogAmount3(pos, SpotPos); //fog attenuates light.
		
		// IBL
		#ifdef USE_PK_IBL
			vec3 ibl = BRDF_IBL(m,n,dir, coloredAO);
		#endif
		
		// spot
		vec3 spot = spotBRDF * FG * spotFalloff;
		
		// sun
		vec3 sun = sunBRDF; 
		
		vec3 totalLight = ambient + spot + sun;
		
		#ifdef USE_PK_IBL
			totalLight += ibl;
		#endif
		
		return totalLight;
	}
	
	
	vec3 cycle(vec3 c, float s) {
		return vec3(0.5)+0.5*vec3(cos(s*Cycles+c.x),cos(s*Cycles+c.y),cos(s*Cycles+c.z));
	}
	
	vec3 getColor() {
		
		orbitTrap.w = sqrt(orbitTrap.w);
		
		vec3 orbitColor;
		if (CycleColors) {
			orbitColor = cycle(X.xyz,orbitTrap.x)*X.w*orbitTrap.x +
			cycle(Y.xyz,orbitTrap.y)*Y.w*orbitTrap.y +
			cycle(Z.xyz,orbitTrap.z)*Z.w*orbitTrap.z +
			cycle(R.xyz,orbitTrap.w)*R.w*orbitTrap.w;
		}
		
		else {
			orbitColor = X.xyz*X.w*orbitTrap.x +
			Y.xyz*Y.w*orbitTrap.y +
			Z.xyz*Z.w*orbitTrap.z +
			R.xyz*R.w*orbitTrap.w;
		}
		
		vec3 color = mix(BaseColor, 3.0*orbitColor,  OrbitStrength);
		return color;
	}
	
	#ifdef  providesColor
		vec3 baseColor(vec3 point, vec3 normal);
	#endif
	
	vec3 colorBase = vec3(0.0,0.0,0.0);
	float epsModified = 0.0;
	bool hitlight = false;
	
	// ------------------------------------------------------ Trace -----------------------------------------
	
	
	vec3 splitToChannels(float f){
		vec3 r = 0.0;
		int f24 = int(clamp(f, 0. ,1.) * pow(2,24)); // Cast to 24b size int -> 0 -> 2^24
		
		r.x = mod(f24, pow(2,24))/pow(2,16); // 24bit comp 0-255 range
		r.y = mod(f24, pow(2,16))/pow(2,8); // 16b comp 0-255 range
		r.z = mod(f24, pow(2,8))/pow(2,0); // 8b comp, 0-255 range
		
		vec3 temp = r.x * pow(2,16) + r.y * pow(2,8) + r.z * pow(2,0);
		return r/pow(2,8);
	}
	
	/*	
	struct SColor{
	vec3 RGB; // RGB
	float A; // Alpha
	float Z; // Depth
	vec3 Tag; // RGB Color used for tagging layers
	bool DoPost; // Whether it should postprocess the pixel
	}
	*/
	
	// 1.0.8 mod	
	//vec3 trace(inout SRay Ray, inout vec3 hitNormal, inout float glow) {	
	SColor trace(inout SRay Ray, inout vec3 hitNormal, inout float glow) {
		
		SColor color = SColor(vec3(0,0,0),1.,0.,vec3(0,0,0),true);
		glow=1000.0;
		vec3 hit = SRCurrentPt(Ray); //from+dir*totalDist;
		orbitTrap = vec4(10000.0);
		
		floorDist = 0.0;
		float dist = 0.0;
		int steps;
		
		bool hitfractal = false;
		bool hitfloor = false;
		bool hitlight = false;
		bool hitbg = false;
		
		vec3 tag;
		float depth;
		float alpha;
		
		colorBase = vec3(0.);
		vec3 hitColor = vec3(0.);
		vec4 background = backgroundColor(Ray.Direction,gl_FragCoord.xy * PixelScale);
		
		
		// We will adjust the minimum distance based on the current zoom
		float eps = minDist;
		float lightde=0.;
		
		vec3 p = SRCurrentPt(Ray);
		float ldist = min(DEF(p) * FudgeFactor, DElight(p,SpotPos));
		ldist *= (Dither*rand(Ray.Direction.xy))+(1.0-Dither);
		SRAdvance(Ray, ldist);
		
		for (steps=0; steps < MaxRaySteps; steps++) {
			vec3 p = SRCurrentPt(Ray); //from + totalDist * direction;
			//dist = DEF(p) * FudgeFactor;
			
			dist = lightde = DElight(p, SpotPos);
			dist = min(dist, DEF(p) * FudgeFactor);
			glow = min(lightde,glow);
			
			SRAdvance(Ray, dist);
			epsModified = max(MIN_EPS, Ray.Pos*eps* FudgeFactor);
			
			if (dist < epsModified) {
				// move back
				SRAdvance(Ray, dist-2.5*epsModified); //totalDist -= (epsModified-dist);
				//float l = ldist*(epsModifiet-dist)/(ldist-1.*dist); SRAdvance(Ray, -ldist);
				break;
			}
			
			if (Ray.Pos > MaxDistance) {
				fSteps -= (Ray.Pos-MaxDistance)/dist;
				break;
			}
			
			//vec3 baseColor(vec3 point, vec3 normal);
			//ldist = dist;
		}
		
		// Event: hit light
		//if(dist==glow) hitlight=true;
		
		// Event:  hit floor
		if (FloorEnable && dist == floorDist * FudgeFactor) hitfloor = true;
		
		// Event: Did not hit anything, not sure wif we need it
		if (steps == MaxRaySteps) orbitTrap = vec4(0.0);
		
		float shadowStrength = 0.0;
		float stepFactor = clamp((fSteps)/float(GlowMax),0.0,1.0);
		
		if (dist < epsModified) {
			// We hit something, or reached MaxRaySteps
			hit = SRCurrentPt(Ray); //from + totalDist * direction;
			float ao = AO.w * stepFactor;
			
			// Event: hit light
			if (dist == lightde) {
				hitColor = Spot.xyz * Spot.w / (SpotSize+0.01);
				
				#ifdef TAG_LAYERS
					tag = tagLight;
				#endif
				
				// Originally 1.07
				// return Spot.xyz* Spot.w/(SpotSize+0.01);
				color.RGB = hitColor;
				color.DoPost = false;
				return color;
			} else 
			
			if (hitfloor){
				
				hitColor = FloorColor;
				hitNormal = floorNormal;
				//if (dot(hitNormal,Ray.Direction)>0.0)
				if(dot(floorNormal,p)-FloorHeight<0.) hitNormal *=-1.0;
				
				// hitColor += Glow.xyz * stepFactor * Glow.w*(1.0-shadowStrength);
				
				#ifdef TAG_LAYERS
					tag = tagFloor;
				#endif
			
			} else {	
				
				hitNormal = normal(hit-NormalBackStep*epsModified*Ray.Direction, epsModified); // *normalE*epsModified/eps*
				
				#ifdef TAG_LAYERS
					tag = tagFractal;
				#endif
				
				hitfractal = true;
			}	
			
			#ifdef providesColor
				hitColor = mix(BaseColor,  baseColor(hit,hitNormal),  OrbitStrength);
			
			#else
				hitColor = getColor();
				
			#endif
			
			/* 1.0.9 mod - I think Gamma was applied twice so removing that
			#ifndef linearGamma
			//	hitColor = pow(clamp(hitColor,0.0,1.0),vec3(Gamma));
			#endif
			*/
			
			if (hitfloor){
				
				if (FloorDebug) {
					vec3 proj = hit - dot(hit, FloorNormal)*hit;
					if (mod(proj.x,1.0)<0.02 || mod(proj.y,1.0)<0.02)
					hitColor  = vec3(0.0);
					if (mod(proj.x,1.0)<0.01 || mod(proj.y,1.0)<0.01)
					hitColor = vec3(1.0);
				color.DoPost = false;
				} else {
					hitColor = FloorColor;
				}
				
				hitColor = mix(background, hitColor, FloorColor.w);
				//		shadowStrength = mix(1.0, shadow(hit+hitNormal*epsModified,epsModified),Shadow);
				//		hitColor*=shadowStrength;
				
				if (FloorMatte){ 
					alpha = float(hitColor);
				hitColor = (0.,0.,0.);
				} else {
					alpha = 1; //float(hitColor);
					//hitColor = (0.,0.,0.);
				}
			}
			
			
			//hitColor = mix(hitColor, AO.xyz ,ao);	
			if (DetailAO < 0.0) ao = ambientOcclusion(hit, hitNormal);
			hitColor = lighting(hitNormal, hitfractal ? ObjMat : FloorMat,  hit,  Ray.Direction, epsModified, shadowStrength, ao);
			
			
			// OpenGL GL_EXP2 like fog
			
			/*
			float f = Ray.Pos;//totalDist;
			//hitColor = mix(hitColor, backColor, 1.0-exp(-pow(Fog,4.0)*f*f));
			hitColor = mix(hitColor, backColor, 1.0-exp(-pow(Fog/FogExp,4.0)*pow(f*f,FogExp)));
			*/
			
		}
		
		else {
			// Event: Hit background
			color.DoPost = false; // exclude bg from post processing
			vec4 background = backgroundColor(Ray.Direction,gl_FragCoord.xy * PixelScale);
			hitColor = background.xyz;
			/*		vec2 pos = ;
			vec2 pixelsiz=vec2(dFdx(pos.x),dFdy(pos.y));
			vec4 tex = texture2D(frontbuffer, pos);
			vec3 c = tex.xyz/tex.a;
			*/
			
			#ifdef TAG_LAYERS
				tag = tagBackground;
			#endif
			
			//hitColor +=Glow.xyz*stepFactor* Glow.w*(1.0-shadowStrength);
			// Debug Sun
			if (DebugSun) {
				vec3 SunDir =  normalize(vec3(sin(SunDir.x*3.1415)*cos(SunDir.y*3.1415/2.0), sin(SunDir.y*3.1415/2.0)*sin(SunDir.x*3.1415), cos(SunDir.x*3.1415)));
				if (dot(SunDir,normalize(Ray.Direction))>0.9) hitColor= vec3(100.,0.,0.);
			}
			
			hitNormal = vec3(0.0);
			Ray.Pos = MaxDistance;//1000.;//if nothing hit return a big value. This is necesary for the fog.
		}
		
		// --------------- Depth
		
		#ifdef DEPTH_PASS
			int dexp = 2;
			float dsmin = 1 / pow(dexp,24);
			float ds = (1-DepthScale) * pow(dexp, DepthScalePower-2);
			depth = 1 - clamp(((Ray.Pos + DepthOffset) / length(Ray.Direction))*ds, dsmin, MaxDistance)/MaxDistance + DepthOffsetPost;
			
			vec3 depthColor = MultiChannelDepth ? splitToChannels(depth) : vec3(depth);
		#endif
		
		// reflections need to be 0 for this to work
		gl_FragDepth = ((1000.0 / (1000.0 - 0.00001)) +
		(1000.0 * 0.00001 / (0.00001 - 1000.0)) /
		clamp(Ray.Pos/length(Ray.Direction), 0.00001, 1000.0));
		
		// --------------- Final
		
		color.RGB = hitColor;
		color.Z = depth;
		color.A = alpha;
		color.Tag = tag;
		
		#ifdef TAG_LAYERS
			if(OutputTag) {
				color.RGB = tag;
				color.DoPost = false; // exclude tags pass from post processing
			}
			
		#endif
		
		#ifdef DEPTH_PASS
			if(OutputDepth){ 
				color.RGB = depth;
				color.DoPost = false; // exclude depth pass from post processing
			}
			
		#endif
		
		return color;
	}
	
	// ---------------------------------------------------- Clouds --------------------------------------------
	
	#ifdef USE_IQ_CLOUDS
		#group IQ_Clouds
		uniform float mtime; slider[-180.00,0,180]
		
		//how large should clouds appear (larger clouds render faster)
		uniform float CloudScale; slider[0.1,1.0,10.0]
		//how flat are they
		uniform float CloudFlatness; slider[0.0,0.0,0.9]
		//how high do the clouds go
		uniform float CloudTops; slider[-10.0,0.0,10.0]
		//and how low
		uniform float CloudBase; slider[-10.0,-1.0,10.0]
		//how thick
		uniform float CloudDensity; slider[0.0,0.0,1.0]
		//are they smooth or rough
		uniform float CloudRoughness; slider[0.0,1.0,2.0]
		//how much light contrast do they create
		uniform float CloudContrast; slider[0.0,1.0,10.0]
		//what is the base color
		uniform vec3 CloudColor; color[0.5,0.5,0.5]
		//and the colfor of light hitting them (posLight doesn't have a color??)
		uniform vec3 SunLightColor; color[0.75,0.75,0.75]
		//wind direction
		uniform vec3 WindDir; slider[(-1.0,-1.0,-1.0),(0.0,1.0,0.0),(1.0,1.0,1.0)]
		//wind speed
		uniform float WindSpeed; slider[0.0,1.0,2.0]
		
		#ifdef USE_IQ_CLOUDS_MULTILAYER
			#group IQ_Clouds High
			//how large should clouds appear (larger clouds render faster)
			uniform float HCloudScale; slider[0.1,1.0,10.0]
			//how flat are they
			uniform float HCloudFlatness; slider[0.0,0.5,0.9]
			//how high do the clouds go
			uniform float HCloudTops; slider[-10.0,2.0,10.0]
			//and how low
			uniform float HCloudBase; slider[-10.0,1.0,10.0]
			//how thick
			uniform float HCloudDensity; slider[0.0,0.0,1.0]
			//are they smooth or rough
			uniform float HCloudRoughness; slider[0.0,1.0,2.0]
			//how much light contrast do they create
			uniform float HCloudContrast; slider[0.0,1.0,10.0]
			//what is the base color
			uniform vec3 HCloudColor; color[0.5,0.5,0.5]
			//wind direction
			uniform vec3 HCloudWindDir; slider[(-1.0,-1.0,-1.0),(0.0,1.0,0.0),(1.0,1.0,1.0)]
			//wind speed
			uniform float HCloudWindSpeed; slider[0.0,1.0,2.0]
		#endif
		
		float rand(vec3 co){
			// implementation found at: lumina.sourceforge.net/Tutorials/Noise.html
			return fract(sin(dot(co*0.123,vec3(12.9898,78.233,112.166))) * 43758.5453);
		}
		
		float cnoyz(vec3 co){
			vec3 d=smoothstep(0.0,1.0,fract(co));
			co=floor(co);
			const vec2 v=vec2(1.0,0.0);
			return mix(mix(mix(rand(co),rand(co+v.xyy),d.x),
			mix(rand(co+v.yxy),rand(co+v.xxy),d.x),
			d.y),mix(mix(rand(co+v.yyx),rand(co+v.xyx),d.x),
			mix(rand(co+v.yxx),rand(co+v.xxx),d.x),d.y),d.z);
		}
		
		float cloudDensity( in vec3 p , in vec3 cloudDir, in float t, float cBase, float cTops, float cDensity, float cScale, float cRoughness, float cFlatness)
		{
			vec3 q=p/cScale;
			q/=vec3(1.0)-cloudDir*cFlatness;
			float f=0.0,a=0.5;
			for(int i=0;i<5;i++){
				f+= a*cnoyz( q );
				q = q*2.03;
				a = a*0.5;
				if(i>6-int(5.0*t/MaxDistance))break;
			}
			
			float y=dot(p,cloudDir);
			float cDen=1.0 - 2.0*abs(0.5*(cTops+cBase)-y)/(cTops-cBase);
			return clamp( cDensity * cDen - cRoughness * f, 0.0, 1.0 );
		}
		
		vec4 integrateClouds( in vec4 sum, in float dif, in float den, in float t ){
			// lighting
			vec3 lin = CloudColor*1.3 + SunLightColor*dif*CloudContrast; // color from the clouds gui
			//vec3 lin = CloudColor*1.3 + SunLight*dif*CloudContrast; // color of the sun
			
			vec4 col = vec4( mix( 1.15*vec3(1.0,0.95,0.8), vec3(0.0), den ), den );
			col.xyz *= lin;
			col=clamp(col,0.0,1.0);
			// col.xyz = mix( col.xyz, bgcol, 1.0-exp(-0.0025*t*t) );
			// front to back blending
			col.a *= mix(0.9,0.0,clamp(t*t/(MaxDistance*MaxDistance),0.0,1.0));
			col.rgb *= col.a;
			return sum + col*(1.0-sum.a);
		}
		
		vec4 clouds(vec3 p0, vec3 p1,  float cBase, float cTops, float cDensity, float cScale, float cRoughness, float cFlatness, vec3 cWindDir, float cWind){
			//vec3 ro=p0 + (WindSpeed*(WindDir*mtime)),rd=normalize(p1-p0),cloudDir=normalize(HF_Dir);
			vec3 ro=p0 + (cWind*(cWindDir*mtime)),rd=normalize(p1-p0),cloudDir=normalize(HF_Dir);
			
			vec4 sum = vec4(0.0);
			float t=0.1*cScale*rand(vec3(gl_FragCoord.xyy+float(subframe)*30.0)),maxT=length(p1-p0);
			bool goingUp=(dot(rd,cloudDir)>0.0);
			while(t<maxT) {
				vec3  pos = ro + t*rd;
				if((!goingUp && dot(pos,cloudDir)<cBase) || (goingUp && dot(pos,cloudDir)>cTops) || sum.a > 0.99) break;
				float den = cloudDensity(pos,cloudDir,t, cBase, cTops, cDensity, cScale,  cRoughness, cFlatness);
				if( den>0.01 ) {
					vec3 sundir=normalize(SpotPos-pos);
					float dif = clamp((den - cloudDensity(pos+0.3*cScale*sundir,cloudDir,t, cBase, cTops, cDensity, cScale,  cRoughness, cFlatness))/0.6, -1.0, 1.0 );
					sum = integrateClouds( sum, dif, den, t );
				}
				
				t += 0.1*cScale+0.02*cScale*t;
			}
			
			return clamp(sum,0.0,1.0);
		}
		
	#endif
	
	#ifdef USE_KN_VOLUMETRIC
		float length2(vec3 p){ return dot(p,p);}
		
		//See: https://www.shadertoy.com/view/Xdf3zB
		//Good for reflatively homogeneous media and for dark ambiance.
		//This uses importance sampling wrt distance to light source. Looks good when anisotropy is small but noisy otherwise.
		//Should use multipe importance sampling wrt:
		// -distance to light (done)
		// -Distance: we don't need details far away. (I don't think this is necessary in this particular case)
		// -Extinction: when fog density is high enought.
		// -Anisotropy: I gess this is as important as distance to light because somehow it have the inverse effect.
		//ToDo2: Take fog color into account
		//ToDo3: correct the use of pseudo-random number generator. (any help?)
		int fogstrata=0;
		vec3 ptLightGlow1(vec3 P0, vec3 P1,vec3 lightPos, vec4 lightColor){
			//Values used for importance sampling
			float A=length2(P0-lightPos);
			float B=dot(P1-P0, P0-lightPos);
			float C=length2(P1-P0);
			float Delta=sqrt(A*C-B*B);
			//get sample point using importance sampling.
			float x=(rand(viewCoord+vec2(1.6183+float(subframe+fogstrata)))+float(fogstrata))*1./float(HF_FogIter); fogstrata++;
			float atanB=atan(B/Delta), atanBC=atan((B+C)/Delta);//B+C=dot((P1-P0)*(P1-LightPos))
			float PDF_NF=(atanBC-atanB)/Delta;//Normalization factor due to division by the PDF.
			float t=(tan(mix(atanB,atanBC,x))*Delta-B)/C;//when x is close to 0. there are black dots!
			t=clamp(t,0.00001,1.); //This is very strange: I have to clamp this way to get rid from black dots. When doing clamp(t,0.,1.) I get much more black dots.
			vec3 Pt=P0+t*(P1-P0);//this is the sample point
			float extP0t=fogAmount(P0,Pt);//Transmittance between "eye" to sample
			float extLt=fogAmount(Pt,lightPos);//Transmittance between sample and light source.
			//Compute fog density at sample (redundent with fogAmount(). Should do something about it)
			vec3 hfdir=normalize(HF_Dir);
			float Ty=HF_Fallof*(dot(Pt,hfdir)-HF_Offset);
			float densiTy=HF_Intensity*exp1(-Ty)+HF_Const; //densiTy=clamp(densiTy,0.,100.);
			//Shadow
			float shadowStrength=1.;
			if(HF_CastShadow)
			shadowStrength = 1.-shadow(Pt, lightPos, 0.001);
			//Anisotopy factor
			float cosTheta=dot(normalize(Pt-P0),normalize(lightPos-Pt));//cos(theta): the angle between light direction and view direction at the sample
			float denom = 1. + HF_Anisotropy.x*HF_Anisotropy.x - 2.*HF_Anisotropy.x*cosTheta;
			float Anie=(1.-HF_Anisotropy.x*HF_Anisotropy.x)/sqrt(denom*denom*denom);//Dropped the 1./(4.*PI)* factor. Well This renderer is not meant to be physically accurate after all. :o)
			//Scattering at sample point
			float samVal=extP0t*extLt*densiTy*Anie*shadowStrength*sqrt(C);//sqrt(C) is the length of the path.
			//Return scattered light
			//samVal=max(0.,samVal);
			return lightColor.xyz* lightColor.w*PDF_NF*samVal*HF_Scatter;
		}
		
		vec3 ptLightGlow3(vec3 P0, vec3 P1){
			
			//Values used for importance sampling
			float A=length2(P0-SpotPos);
			float B=dot(P1-P0, P0-SpotPos);
			float C=length2(P1-P0);
			float Delta=sqrt(A*C-B*B);
			
			//get sample point using importance sampling.
			float x=(rand(viewCoord+vec2(1.6183+float(subframe+fogstrata)))+float(fogstrata))*1./float(HF_FogIter); fogstrata++;
			float atanB=atan(B/Delta), atanBC=atan((B+C)/Delta);//B+C=dot((P1-P0)*(P1-LightPos))
			float PDF_NF=(atanBC-atanB)/Delta;//Normalization factor due to division by the PDF.
			float t=(tan(mix(atanB,atanBC,x))*Delta-B)/C;//when x is close to 0. there are black dots!
			t=clamp(t,0.00001,1.); //This is very strange: I have to clamp this way to get rid from black dots. When doing clamp(t,0.,1.) I get much more black dots.
			vec3 Pt=P0+t*(P1-P0);//this is the sample point
			vec3 extP0t=fogAmount3(P0,Pt);//Transmittance between "eye" to sample
			vec3 extLt=fogAmount3(Pt,SpotPos);//Transmittance between sample and light source.
			
			//Compute fog density at sample (redundent with fogAmount(). Should do something about it)
			vec3 hfdir=normalize(HF_Dir);
			float Ty=HF_Fallof*(dot(Pt,hfdir)-HF_Offset);
			vec3 densiTy=HF_Color.rgb*HF_Intensity*exp1(-Ty)+HF_Const; //densiTy=clamp(densiTy,0.,100.);
			
			//Shadow
			float shadowStrength=1.;
			if(HF_CastShadow)
			shadowStrength = 1.-shadow(Pt, SpotPos, 0.001);
			
			//Anisotopy factor
			float cosTheta=dot(normalize(Pt-P0),normalize(SpotPos-Pt));//cos(theta): the angle between light direction and view direction at the sample
			vec3 denom = 1. + HF_Anisotropy*HF_Anisotropy - 2.*HF_Anisotropy*cosTheta;
			vec3 Anie=(1.-HF_Anisotropy*HF_Anisotropy)/sqrt(denom*denom*denom);//Dropped the 1./(4.*PI)* factor. Well This renderer is not meant to be physically accurate after all. :o)
			//Scattering at sample point
			vec3 samVal=extP0t*extLt*densiTy*Anie*shadowStrength*sqrt(C);//sqrt(C) is the length of the path.
			//Return scattered light
			//samVal=max(0.,samVal);
			return Spot.xyz*Spot.w*PDF_NF*samVal*HF_Scatter;
		}
		
	#endif
	
	// ---------------------------------------------------- Color --------------------------------------------
	
	//vec3 color(SRay Ray) 
	
	/*
	// Just for reference
	struct SColor{
	vec3 RGB; // RGB
	float A; // Alpha
	float Z; // Depth
	vec3 Tag; // RGB Color used for tagging layers
	bool DoPost; // Whether it should postprocess the pixel
	}
	
	;
	*/
	
	SColor color(SRay Ray) {
		float glow=0.5;
		vec3 hitNormal = vec3(0.0);
		vec3 col = vec3(0.0);
		vec3 col0 = vec3(0.0);
		vec3 RWeight=vec3(1);
		
		SColor Color;
		// Reflections loop
		for(int i=0; i<= ReflectionsNumber; i++){
			vec3 prevPos = SRCurrentPt(Ray);
			
			// ----> TRACE <------
			
			Color = trace(Ray, hitNormal, glow);
			col0 = Color.RGB;
			
			vec3 curPos= SRCurrentPt(Ray);
			vec3 FG = fogAmount3(prevPos, curPos);//get fog amount
			col0=mix(HF_Color.rgb*HF_Color.w,col0,FG);//modify color
			col0+=ptLightGlow(max(0.,glow),Spot);
			
			#ifdef USE_KN_VOLUMETRIC
				if(HF_Scatter*(HF_Intensity+HF_Const)>0.){
					vec3 colfogglow=vec3(0.);
					for(int j=0;j<HF_FogIter;j++){
						colfogglow+=ptLightGlow3(prevPos, curPos);
					}
					
					col0+=colfogglow*1./float(HF_FogIter);
					//ptLightGlow1(prevPos, curPos);
				}
				
			#endif
			
			#ifdef USE_IQ_CLOUDS
				//  base layer
				vec4 clds = clouds(prevPos, curPos, CloudBase, CloudTops, CloudDensity, CloudScale, CloudRoughness, CloudFlatness, WindDir, WindSpeed);
				col0=col0*(1.0-clds.w)+clds.rgb;
			#endif
			
			// top layer
			#ifdef USE_IQ_CLOUDS_MULTILAYER
				clds = clouds(prevPos, curPos, HCloudBase, HCloudTops, HCloudDensity, HCloudScale, HCloudRoughness, HCloudFlatness, HCloudWindDir, HCloudWindSpeed);
				col0 = col0*(1.0-clds.w)+clds.rgb;
				//ToDo windspeed winddir
			#endif
			
			
			// Reflections
			col+=col0*RWeight;
			RWeight *= Reflection * FG;//modify current opacity
			
			if (hitNormal == vec3(0.0) || dot(RWeight,RWeight)<0.0001 || hitlight || Ray.Pos>=MaxDistance) {
				
				
				//nothing hit or light hit or reflected light is too small
				break;
			}
			
			
			// reflect the ray
			Ray = SRReflect(Ray, hitNormal, 0.*minDist);
		}
		
		Color.RGB = max(vec3(0.),col);
		
		//Sometimes col<BigNegativeValue  ->black dots. Why? I don't know :-/. Solved by Eiffie & Syntopia See: http://www.fractalforums.com/fragmentarium/updating-of-de-raytracer/msg81003/#msg81003 . I keep it just in case .
		
		return Color;
	}
		