#include "Classic-Noise.frag"

#group Tranform
// Added global transformations to the space

// Translate
uniform vec3 sTranslate; slider[(-10,-10,-10),(0,0,0),(10,10,10)]

// Rotate
uniform vec3 sRotate; slider[(-180,-180,-180),(0,0,0),(180,180,180)]

// Scale
uniform vec3 sScale; slider[(0.001,0.001,0.001),(1,1,1),(10,10,10)]

// Twist
uniform float sTwist; slider[-2,0,2]

// Stereographic projection
uniform bool sStereo; checkbox[false]
// Mul = 2 is standard stereographic projection
uniform float sStereoF; slider[0,2,3]
// Stereo transform 1
uniform float sStereoD1; slider[0,1,5]
// Stereo transform 2
uniform float sStereoD2; slider[0,2,5]
// Stereo offset
uniform vec3 sStereoOffset; slider[(-2,-2,-2),(0,0,0),(2,2,2)]
// Amount of stereographic projection
uniform float sStereoMix;  slider[0.0,1,1.0]

#group Clipping
// Allows for slicing the fractal and rendering only a box-limited volume
// Box may be sclaed/rotated/translated


// Clip scene contents to box
uniform bool clipToBox; checkbox[false]
// Use that to see the box
uniform bool showClippingBox; checkbox[false]

// Clipping Box size
uniform vec3 cBoxSize; slider[(0,0,0),(5,5,5),(10,10,10)]

// Clipping Box Translate
uniform vec3 cBoxTranslate; slider[(-10,-10,-10),(0,0,0),(10,10,10)]

// Clipping Box Rotate
uniform vec3 cBoxRotate; slider[(-180,-180,-180),(0,0,0),(180,180,180)]

#group Noise
uniform float nTime; slider[0.00,1.0,4.00]

// Noise Scale
uniform vec3 nScale; slider[(0,0,0),(1,1,1),(5,5,5)]

// Noise Amplitude
uniform vec3 nAmplitude; slider[(0,0,0),(0,0,0),(2.5,2.5,2.5)]

// -------------------------------- Functions -------------------------------

// Stereo mapping
vec4 stereographic3Sphere(vec3 p) {
	float n = dot(p,p)+sStereoD1;
	return vec4(sStereoF*p,n-sStereoD2)/n;
}

// Noise
vec3 spaceNoise (vec3 p){
	if (nAmplitude.x > 0)	p += nAmplitude.x*vec3(cnoise(vec4(p*nScale.x,nTime)));
	if (nAmplitude.y > 0)	p += nAmplitude.y*0.3*vec3(cnoise( vec4(p*nScale.y*3,nTime)));
	if (nAmplitude.z > 0)	p += nAmplitude.z*0.02*vec3(cnoise( vec4(p*nScale.z*20,nTime)));
	
	return p;
}

vec3 stereoMap (vec3 p){
	if (sStereo){
		float rr = length(p);
		vec4 sp = stereographic3Sphere(p);
		sp.xyz += sStereoOffset;
		//Inverse stereographic projection of pos: z4 lies onto the unit 3-sphere centered at 0.
		//sp = vec4(2.*z,-1.+rr*rr)*1./(1.+rr*rr);
		p = sp * sStereoMix + p * (1-sStereoMix);
	}
	return p;
}

// Returns DE to a box
float DEBox(vec3 p, vec3 box){
	vec3 boxd = abs(p) -  box;
	return min(max(boxd.x,max(boxd.y,boxd.z)),0.0) + length(max(boxd,0.0));
}

// -------------------------------- IQ Operators -------------------------------

vec3 opTranslate(vec3 p, vec3 d){
	mat4 m = translate(d);
	//	vec3 p1 = inverse(m)*vec4(p.x, p.y, p.z,1); // slow way
	vec3 p1 = m * vec4(p.x, p.y, p.z,1) * -1; //-1 is for settings compatibility
	
	return p1;
}

vec3 opRotate(vec3 p, vec3 v, float angle){
	mat4 m = rotationMatrix(normalize(v),angle);
	//	vec3 p1 = inverse(m)*vec4(p.x, p.y, p.z,1);
	vec3 p1 = m * vec4(p.x, p.y, p.z,1) * -1;
	return p1;
}

vec3 opScale( vec3 p, float s){
	return (p / s);
}

vec3 opNUScale( vec3 p, vec3 s){
	return (p / s);
}

vec3 opTwist( vec3 p, float tw)
{
	float c = cos(tw*p.y);
	float s = sin(tw*p.y);
	mat2  m = mat2(c,-s,s,c);
	vec3  q = vec3(p.xz*m,p.y);
	return q;
}
