// Output generated from file: //production/Public/Fractals/Fragmentarium/Frags/Dev_3_Matchmoving.frag
// Created: So sie 1 00:02:33 2015
// Based on original classic Quaternion bundled with Fragmentarium
// Modifications added by Patryk Kizny / kizny.com

// 2015/06/15
// - Added Box folding
// - Added Sphere Folding
// - Added Volume clipping
// - Added general fractal repositioning

#info 4D Quaternion Julia DE + Box Folding + Sphere Folding

// Flags
#define providesInit
#define USE_EIFFIE_SHADOW
#define PERFECT_DE
#define USE_PK_IBL
#define USE_PK_BACKGROUND
//#define USE_IQ_CLOUDS
//#define USE_IQ_CLOUDS_MULTILAYER // Multilayer Clouds by PK
//#define USE_KN_VOLUMETRIC

#define TAG_LAYERS // Uncomment this to output alpha pass
#define DEPTH_PASS // Uncomment this to output depth pass

uniform float time;

// Core
//#include "ExternalCamera-0.1.frag"
#include "Presets-Camera-Z-Up.frag"
#include "MathUtils.frag"
#include "Quaternion-Maths.frag"
#include "SpaceTransforms.frag"
#include "DE-PK-1.0.11.frag"


#group 4D Quaternion Julia
// Number of fractal iterations.
uniform int QTIterations;  slider[0,16,100]

// Breakout Distance
uniform float QTThreshold; slider[0,10,100]

// Quaterion Constant C
uniform vec4 QTC; slider[(-1,-1,-1,-1),(0.18,0.88,0.24,0.16),(1,1,1,1)]

// Iterative Rotation
uniform vec3 rotVector; slider[(0,0,0),(0,0,1),(1,1,1)]
uniform float rotAngle; slider[0.00,0,180]

// Iterative Scaling
uniform float Scale;  slider[-3.0,1,3.0]

#group box Folding
// Added box folding by Patryk Kizny
// Sphere folding does not seem to work for this fractal

uniform int foldBoxStartIt; slider[0,0,256]
uniform int foldBoxIts; slider[0,0,16]

uniform vec3 foldBoxLimit; slider[(-5,-5,-5),(0,0,0),(5,5,5)]
uniform vec3 foldBoxValue; slider[(-5,-5,-5),(2,2,2),(5,5,5)]
uniform vec3 foldBoxOffset; slider[(-5,-5,-5),(0.,0.,0.),(5,5,5)]

#group sphere Folding
uniform int foldSphereStartIt; slider[0,0,256]
uniform int foldSphereIts; slider[0,0,16]
uniform float foldSphereMinRad;  slider[0,0.25,2.0]

// ---------------------------------------- Init -------------------------------------------

mat3 rot; // QT
void init() {
	rot = rotationMatrix3(normalize(rotVector),rotAngle);
}

// ---------------------------------------- Quaternion Modified -------------------------------------------

vec4 scale = vec4(Scale, Scale, Scale, abs(Scale)) / foldSphereMinRad;

// DE implementation here based on Syntopia + Subblue + Keenan Crane
// http://www.subblue.com/blog/2009/9/20/quaternion_julia
// Added box folding, sphere folding and iterative rotation

float DEQT(vec3 pos) {
	
	vec4 dp = vec4(1.0, 0.0,0.0,0.0);
	vec4 p = vec4(pos, 0.0);

	for (int i = 0; i <QTIterations; i++){
		
		// Quaternion interations
		// Syntopia's compressed version
		//	dp = 2.0* vec4(p.x*dp.x-dot(p.yzw, dp.yzw), p.x*dp.yzw+dp.x*p.yzw+cross(p.yzw, dp.yzw));
		//	p = vec4(p.x*p.x-dot(p.yzw, p.yzw), vec3(2.0*p.x*p.yzw)) + C;
		
		// Experimental, generates artifacts
		//	dp = QTDerCons * qMult(p,dp);
		dp = 2.0 * qMult(p,dp);
		p = qSq(p) + QTC;
		
		// rotation
		p.xyz *= rot;
		
		// scaling, experimental, generates artifacts
		// p *= scale;
		
		// Box Folding (v2) - separate axes + value + offset
		if (foldBoxIts > 0){
			if ( i >= foldBoxStartIt && i < foldBoxStartIt + foldBoxIts){
				p.xyz = clamp(p.xyz, -foldBoxLimit, foldBoxLimit) * foldBoxValue - p.xyz + foldBoxOffset;
			}
		}
		
		// Sphere folding
		if(foldSphereIts > 0){
			if ( i >= foldSphereStartIt && i < foldSphereStartIt + foldSphereIts){
				float r2 = dot(p.xyz, p.xyz);
				p *= clamp(max(foldSphereMinRad/r2, foldSphereMinRad), 0.0, 1.0);
			}
		}
		
		float p2 = dot(p,p);
		orbitTrap = min(orbitTrap, abs(vec4(p.xyz,p2)));
		if (p2 > QTThreshold) break;
	}
	
	// DE
	float r = length(p);
	return  0.5 * r * log(r) / length(dp);
}


// ----------------------------------------Main -------------------------------------------

float DE(vec3 z) {

	// get slice volume
	vec3 z1;
	z1 = opRotate(z, cBoxRotate, length(cBoxRotate));
	z1 = opTranslate(z1, cBoxTranslate - sTranslate);
	float deBox = DEBox(z1, cBoxSize);
	
	float deBox2 = DEBox(vec3(z.x+0.25,z.y-0.5,z.z), vec3(2.0,2.0,0.001));


	// apply global transformations
	z = opNUScale(z, sScale);
	z = opTranslate(z, sTranslate);
	z = opRotate(z, normalize(sRotate), length(sRotate));
//	z = opTwist(z,sTwist);

	// Stereo projection
	z = stereoMap(z);
	
	// Noise
	z = spaceNoise(z);

	//  Instantiate
	//	z.xz = mod(z.xz,5) -0.0; //+vec2(0.1); //spacingXY
	
	//get fractal
	float deQT = DEQT(z);
	// float deQT = length(z)-1; //sphere
 
	// return plain fractal
	float DE = deQT;
	
	// clip to volume
	if (clipToBox){
		DE = max(deQT, deBox);
	}
	// disply cutting volume
	if (showClippingBox) {
		DE = min(deQT, deBox);
	}
//	DE = min(DE, deBox2);
	return DE;

}

// ---------------------------------------- Presets -------------------------------------------





#preset default
sTranslate = 0,0,0
sRotate = 0,0,0
sScale = 0.7,0.7,0.7
sTwist = 0
sStereo = false
sStereoF = 2
sStereoD1 = 1
sStereoD2 = 2
sStereoOffset = 0,0,0
sStereoMix = 1
clipToBox = false Locked
showClippingBox = false Locked
cBoxSize = 1,1,0.01
cBoxTranslate = 0,0,0
cBoxRotate = 0,0,0
nTime = 1
nScale = 1,1,1
nAmplitude = 0,0,0
FOV = 1.35
UseTrueFOV = true
TrueFOV = 115
UseFocalLength = false
FocalLength = 25.6745
FilmGate = 20.25
Eye = -0.562226,-0.731061,0.736054
Ey_ = 0,0,0
Target = 1.33837,0.799971,-1.00855
Targe_ = 0,0,0
FocalPlane = 1
Aperture = 0
InFocusAWidth = 0.5
ApertureNbrSides = 2
ApertureRot = 0
ApStarShaped = false
Gamma = 2
ToneMapping = 3
Exposure = 1
Brightness = 1
Contrast = 1
Saturation = 1
GaussianWeight = 1
AntiAliasScale = 2
Bloom = false
BloomIntensity = 0.25
BloomPow = 2
BloomTaps = 4
Detail = -3.75
DetailAO = -0.5
AutoFocus = true
FudgeFactor = 0.95
MaxRaySteps = 768
MaxDistance = 32
MaxRayStepsDiv = 1.85
Dither = 0.5
NormalBackStep = 1
DepthScale = -0.1
DepthOffset = 0.4
DepthOffsetPost = 0
DepthScalePower = 4
MultiChannelDepth = false
FlatPass = false
tagBackground = 0,0,0
tagLight = 1,0,0
tagFloor = 0,1,0
tagObject = 0,1,1
tagFractal = 1,1,1
AO = 0,0,0,1
AoCorrect = 0
SunLight = 1,1,1,1
SunDir = -0.3,-0.9
DebugSun = false
Glow = 1,1,1,0
GlowMax = 256
Fog = 0.3
FogExp = 2
Reflection = 0,0,0
ReflectionsNumber = 0
BaseColor = 1,0,0
OrbitStrength = 0.1
X = 0,0.666667,0.498039,0.5
Y = 0,0.666667,1,0.5
Z = 1,0,0.498039,0.5
R = 1,1,1,0.5
CycleColors = false
Cycles = 1
FloorNormal = 0,0,1
FloorHeight = 0
HF_Fallof = 0.1
HF_Const = 0
HF_Intensity = 0
HF_Dir = 0,0,1
HF_Offset = 0
HF_Color = 1,1,1,1
HF_Scatter = 0
HF_Anisotropy = 0,0,0
HF_FogIter = 1
HF_CastShadow = false
mtime = 0
CloudScale = 1
CloudFlatness = 0
CloudTops = 0
CloudBase = -1
CloudDensity = 0
CloudRoughness = 1
CloudContrast = 1
CloudColor = 0.5,0.5,0.5
SunLightColor = 0.75,0.75,0.75
WindDir = 0,1,0
WindSpeed = 1
HCloudScale = 1
HCloudFlatness = 0.5
HCloudTops = 2
HCloudBase = 1
HCloudDensity = 0
HCloudRoughness = 1
HCloudContrast = 1
HCloudColor = 0.5,0.5,0.5
HCloudWindDir = 0,1,0
HCloudWindSpeed = 1
QTIterations = 16
QTThreshold = 24
QTC = 0.1,0.85,0.1,0
rotVector = 0,1,0
rotAngle = 0
Scale = 1
foldBoxStartIt = 256 Locked
foldBoxIts = 0 Locked
foldBoxLimit = 0,0,0
foldBoxValue = 2,2,2
foldBoxOffset = 0,0,0
foldSphereStartIt = 256 Locked
foldSphereIts = 3 Locked
foldSphereMinRad = 0.25
DryPass = false
OutputDepth = false
OutputTag = false
fDiffuse = 0.666667,0.666667,0.666667
fDiffExp = 1
fSpecular = 1,1,1,0.75
fSpecularExp = 100
fSpecularMax = 2
oDiffuse = 1,1,1
oDiffExp = 1
oSpecular = 1,1,1,0.2
oSpecularExp = 75
oSpecularMax = 4
AmbientLight = 1,1,1,0.3
AmbientLightMin = 0.1
HardShadow = 1
ShadowSoft = 5
SpotGlow = true
Spot = 1,1,1,0.75
SpotPos = -0.100004,-0.799471,0.55
SpotSize = 0.01
SpotFallOff = 1.5
SpotGlowRad = 0
SpotGlowExp = 0
FloorEnable = true
FloorMatte = false
FloorDebug = false
FloorColor = 0,0,0,1
BgColor = 0.5,0.5,0.5,1
BgGradient = 0.3
BgUseMap = true
BgMapEquirectangular = true
BgMap = C:/Users/Paco/Desktop/Fragmentarium 3Dickulus/Examples/Include/Ditch-River_2k.hdr
BgMapScale = 1,1,1
BgMapOffset = 0,0,0
BgColor = 0.5,0.5,0.5,1
EnvUseBackground = false
EnvMapSpecular = C:/Users/Paco/Desktop/Fragmentarium 3Dickulus/Examples/Include/Ditch-River_2k.hdr
EnvMapDiffuse = C:/Users/Paco/Desktop/Fragmentarium 3Dickulus/Examples/Include/Ditch-River_Env.hdr
EnvUseDiffusionMap = true
EnvLightFactor = 1,1,1
EnvDiffussionFade = 1
EnvSpecularFade = 0.3
Up = 0,0,1
#endpreset

